from . import http_response
from .. import counter
import time
import re
try:
    import ujson as json
except ImportError:
    import json
from rs4.attrdict import CaseInsensitiveDict
from aquests.protocols.http import http_util
from rs4 import jwt

class JWTUser:
    def __init__ (self, claims):
        self.__claims = claims

    @property
    def name (self):
        return self.__claims ["username"]

    def __getattr__ (self, attr):
        return self.__claims.get (attr)

    def __str__ (self):
        return self.name

class http_request:
    response_class = http_response.http_response
    version = "1.1"
    collector = None
    producer = None
    request_count = counter.counter()

    def __init__ (self, *args):
        self.request_number = self.request_count.inc()
        (self.channel, self.request,
         self.command, self.uri, self.version,
         self.header) = args
        self.logger = self.channel.server.server_logger
        self.server_ident = self.channel.server.SERVER_IDENT
        self.body = None
        self.multipart = None
        self.reply_code = 200
        self.reply_message = ""
        self.loadbalance_retry = 0
        self.rbytes = 0
        self.created = time.time ()
        self.gzip_encoded = False
        self.env = None

        self._split_uri = None
        self._headers_cache = None
        self._header_cache = {}
        self._is_stream_ended = False
        self._is_async_streaming = False
        self._is_promise = False
        self._xmlrpc_serialized = False # used by testuitl.client
        self._jwt = None
        self.args = {}

        self.PARAMS = {}
        self.URL = {}
        self.JSON = None
        self.FORM = None
        self.set_log_info ()
        self.make_response ()

    # arguments and parameters -------------------------
    @property
    def JWT (self):
        return self._jwt or self.dejwt ()

    @property
    def DATA (self):
        # FORM or JSON
        return self.dict ()

    @property
    def ARGS (self):
        # all of them
        return self.args

    @property
    def DEFAULT (self):
        try:
            return self.routable ["defaults"]
        except KeyError:
            pass

    def _override_defaults (self, dict):
        if "defaults" in self.routable:
            for k, v in self.routable ["defaults"].items ():
                if k not in dict:
                    dict [k] = v
        return dict

    def set_args (self, args):
        self.args = self._override_defaults (args)

    # processing realted ---------------------------------
    @property
    def current_was (self):
        return self.env ["skitai.was"]

    @property
    def current_app (self):
        return self.env ["wsgi.app"]

    @property
    def routed (self):
        return self.env ["wsgi.routed"]

    @property
    def routable (self):
        return self.env ["wsgi.route_options"]

    @property
    def salt (self):
        return self.current_app.salt

    @property
    def session (self):
        return self.current_was.session

    @property
    def cookie (self):
        return self.current_was.cookie

    @property
    def mbox (self):
        return self.current_was.mbox

    @property
    def salt (self):
        return self.env ["wsgi.app"].salt

    # HTTP protocol constants ----------------------------------
    @property
    def method (self):
        return self.command.upper ()

    @property
    def scheme (self):
        return self.get_scheme ()

    @property
    def headers (self):
        if self._headers_cache:
            return self._headers_cache
        h = CaseInsensitiveDict ()
        for line in self.header:
            k, v = line.split (":", 1)
            h [k] = v.strip ()
        self._headers_cache = h
        return h

    @property
    def payload (self):
        return self.get_body ()

    @property
    def charset (self):
        return self.get_charset ()

    @property
    def content_length (self):
        return self.get_content_length ()

    @property
    def content_type (self):
        return self.get_content_type ()

    @property
    def main_type (self):
        return self.get_main_type ()

    @property
    def sub_type (self):
        return self.get_sub_type ()

    @property
    def user_agent (self):
        return self.get_user_agent ()

    @property
    def remote_addr (self):
        return self.get_remote_addr ()

    @property
    def host (self):
        return self.get_host ()

    @property
    def referer (self):
        return self.get_header ('referer', '')

    @property
    def origin (self):
        return self.get_header ('origin', '')

    @property
    def real_ip (self):
        return self.get_real_ip ()

    @property
    def acceptables (self):
        accept = self.get_header ('accept', '')
        return http_util.parse_multi_params (accept) if accept else {}

    def get_real_ip (self):
        sock_ip = self.get_remote_addr ()
        origin_ips = self.get_header ("X-Forwarded-For")
        if not origin_ips:
            return socke_ip
        return origin_ips.split (",", 1)[0].strip ()

    def get_scheme (self):
        return hasattr (self.channel.server, 'ctx') and "https" or "http"

    def get_charset (self):
        return self.get_attr ("content-type", "charset")

    def get_content_length (self):
        try: return int (self.get_header ("content-length"))
        except: return None

    def get_content_type (self):
        return self.get_header_with_attr ("content-type") [0]

    def get_main_type (self):
        ct = self.get_content_type ()
        if ct is None:
            return
        return ct.split ("/", 1) [0]

    def get_sub_type (self):
        ct = self.get_content_type ()
        if ct is None:
            return
        return ct.split ("/", 1) [1]

    def get_host (self):
        return self.get_header ("host")

    def get_user_agent (self):
        return self.get_header ("user-agent")

    def get_remote_addr (self):
        return self.channel.addr [0]

    def get_real_ip (self):
        ips = self.get_header ("X-Forwarded-For")
        if not ips:
            return self.channel.addr [0]
        return ips.split (",", 1)[0].strip ()

    def acceptable (self, media, strict = True):
        accept = self.get_header ('accept', '')
        if not strict:
            return accept.find ('*/*') != -1
        return accept.find (media) != -1

    def dejwt (self, token = None, salt = None):
        if not token:
            if self._jwt:
                return self._jwt
            token_ = self.get_header ("authorization")
            if not token_ or token_ [:7].lower () != "bearer ":
                self._jwt = {"err": "no bearer token"}
                return self._jwt
            token = token_ [7:]

        try:
            claims = jwt.get_claim (salt or self.salt, token)
        except (TypeError, ValueError):
            claims = {"err": "invalid token"}
        else:
            if claims is None:
                claims = {"err": "invalid signature"}
            else:
                now = time.time ()
                if claims ["exp"] < now:
                    claims = {"err": "token expired"}
                elif "nbf" in claims and claims ["nbf"] > now:
                    claims = {"err": "token not activated yet"}

        self._jwt = claims
        if "username" in claims:
            self.user = JWTUser (claims)
        return claims

    # payload ------------------------------------------------
    def set_multipart (self, dict):
        self.multipart = dict

    def get_multipart (self):
        return self.multipart

    def set_body (self, body):
        self.body = body

    def get_body (self):
        return self.body
    get_payload = get_body

    def json (self, ct = None):
        if not ct:
            ct = self.get_header ('content-type', '')
        if not ct.startswith ("application/json"):
            return
        if self.JSON is not None:
            return self.JSON
        self.JSON = json.loads (self.body.decode ('utf8'))
        return self.JSON

    def form (self, ct = None):
        if self.FORM is not None:
            return self.FORM
        if not ct:
            ct = self.get_header ('content-type', '')
        if self.multipart:
            self.FORM = self.multipart
        elif ct.startswith ("application/x-www-form-urlencoded"):
            self.FORM = http_util.crack_query (self.body)
        return self.FORM

    def dict (self):
        if not self.multipart and not self.body:
            return {}
        ct = self.get_header ('content-type', '')
        maybe = self.json (ct)
        if maybe is None:
            maybe = self.form (ct)
        return maybe or {}

    # logging ---------------------------------------------
    def get_gtxid (self):
        return self.gtxid

    def get_ltxid (self, delta = 1):
        self.ltxid += delta
        return str (self.ltxid)

    def set_log_info (self):
        self.token = None
        self.claim = None
        self.user = None

        self.gtxid = self.get_header ("X-Gtxn-Id")
        if not self.gtxid:
            self.gtxid = "%s-%s-%s" % (
                self.channel.server.hash_id,
                self.channel.channel_number,
                self.request_count
            )
            self.ltxid = 1000
        else:
            self.ltxid = self.get_header ("X-Ltxn-Id")
            if not self.ltxid:
                raise ValueError ("Local txn ID missing")
            self.ltxid = int (self.ltxid) + 1000

    # headers ---------------------------------------------
    def get_raw_header (self):
        return self.header
    get_headers = get_raw_header

    def get_header_with_regex (self, head_reg, group):
        for line in self.header:
            m = head_reg.match (line)
            if m.end() == len(line):
                return head_reg.group (group)
        return ''

    def set_header (self, name, value):
        self.header.append ("%s: %s" % (name, value))

    def get_header (self, header = None, default = None):
        if header is None:
            return self.header
        header = header.lower()
        hc = self._header_cache
        if header not in hc:
            h = header + ':'
            hl = len(h)
            for line in self.header:
                if line [:hl].lower() == h:
                    r = line [hl:].strip ()
                    hc [header] = r
                    return r
            hc [header] = None
            return default
        else:
            return hc[header] is not None and hc[header] or default

    def get_header_params (self, header, default = None):
        v = self.get_header (header, default)
        if v is None:
            return default, {}
        return http_util.parse_params (v)
    get_header_with_attr = get_header_params

    def get_header_noparam (self, header, default = None):
        return self.get_header_params (header, default) [0]

    def get_attr (self, header, attrname = None, default = None):
        value, attrs = self.get_header_params (header, None)
        if not value:
            return default
        if not attrname:
            return attrs
        return attrs.get (attrname, default)

    # publics ------------------------------------------------
    def make_response (self):
        self.response = self.response_class (self)

    def is_promise (self):
        return self._is_promise

    path_regex = re.compile (r'([^;?#]*)(;[^?#]*)?(\?[^#]*)?(#.*)?')
    def split_uri (self):
        if self._split_uri is None:
            m = self.path_regex.match (self.uri)
            if m.end() != len(self.uri):
                raise ValueError("Broken URI")
            else:
                self._split_uri = m.groups()
        return self._split_uri

    def is_private_ip (self, ip = None):
        ip = ip or self.channel.addr [0]
        if ip [:8] in ("127.0.0.", "192.168."):
            return ip
        if ip [:3] == "10.":
            return ip
        if ip [:4] == "172."and 16 <= int (ip [4:].split (".", 1)[0]) < 32:
            return ip

    # channel communicating --------------------------------------
    def set_streaming (self):
        self._is_async_streaming = True

    def is_async_streaming (self):
        return self._is_async_streaming

    def set_stream_ended (self):
        self._is_stream_ended = True

    def is_stream_ended (self):
        return self._is_stream_ended

    def response_finished (self):
        if self.response:
            self.response.request = None

    def xmlrpc_serialized (self):
        # for compat with client request
        return self._xmlrpc_serialized

    def collect_incoming_data (self, data):
        if self.collector:
            self.rbytes += len (data)
            self.collector.collect_incoming_data (data)
        else:
            self.logger.log (
                'dropping %d bytes of incoming request data' % len(data),
                'warn'
                )

    def found_terminator (self):
        if self.collector:
            self.collector.found_terminator()
        else:
            self.logger.log (
                'unexpected end-of-record for incoming request',
                'warn'
                )




#!/usr/bin/env python

import sys
from rs4 import asyncore, asynchat
import re, socket, time, threading, os
from . import http_request
from .. import counter
from aquests.protocols.http import http_util, http_date
from aquests.athreads import threadlib
from skitai import lifetime
from rs4 import producers, compressors
from collections import deque
import signal
import ssl
import skitai
from hashlib import md5
from rs4.psutil.processutil import set_process_name, drop_privileges

if os.name == "posix":
    import psutil
    CPUs = psutil.cpu_count()

PID = {}
ACTIVE_WORKERS = 0
SURVAIL = True
EXITCODE = 0
DEBUG = False


#-------------------------------------------------------------------
# server channel
#-------------------------------------------------------------------
class http_channel (asynchat.async_chat):
    current_request = None
    channel_count = counter.counter ()
    closed = False
    is_rejected = False

    keep_alive = 30
    network_timeout = 30
    zombie_timeout = 30

    fifo_class = deque

    def __init__ (self, server, conn, addr):
        self.conn = conn
        self.server = server
        self.channel_number = http_channel.channel_count.inc ()
        self.request_counter = counter.counter()
        self.bytes_out = counter.counter()
        self.bytes_in = counter.counter()

        self.ac_in_buffer = b''
        self.incoming = []
        self.producer_fifo = self.fifo_class ()
        asyncore.dispatcher.__init__(self, conn)
        self.addr = addr # SHOULD place after asyncore.dispatcher.__init__
        self.set_terminator (b'\r\n\r\n')
        self.in_buffer = b''
        self.creation_time = int (time.time())
        self.event_time = int (time.time())
        self.__history = []
        self.__sendlock = None
        self.producers_attend_to = []
        self.things_die_with = []

    def link_protocol_writer (self):
        self.writable = self._writable_with_protocol
        self.handle_write = self._handle_write_with_protocol

    def _writable_with_protocol (self):
        return self.producer_fifo or (self.current_request and self.current_request.has_sendables ())

    def _handle_write_with_protocol (self):
        data_to_send = self.current_request.data_to_send ()
        written = False
        for data in data_to_send:
            written = True
            self.push (data)
        super ().handle_write ()
        return written

    def use_sendlock (self):
        self.__sendlock  = threading.Lock ()
        self.initiate_send = self._initiate_send_ts

    def get_history (self):
        return self.__history

    def reject (self):
        self.is_rejected = True

    def _initiate_send_ts (self):
        lock = self.__sendlock
        lock.acquire ()
        try:
            asynchat.async_chat.initiate_send (self)
            try: is_working = self.producer_fifo.working ()
            except AttributeError:     is_working = len (self.producer_fifo)
        finally:
            lock.release ()
        if not is_working:
            self.done_request ()

    def initiate_send (self):
        asynchat.async_chat.initiate_send (self)
        try: is_working = self.producer_fifo.working ()
        except AttributeError:     is_working = len (self.producer_fifo)
        if not is_working:
            self.done_request ()

    def readable (self):
        return not self.is_rejected and asynchat.async_chat.readable (self)

    def issent (self):
        return self.bytes_out.as_long ()

    def __repr__ (self):
        ar = asynchat.async_chat.__repr__(self) [1:-1]
        return '<%s channel#: %s-%s requests:%s>' % (
                ar,
                self.server.worker_ident,
                self.channel_number,
                self.request_counter
                )

    def clean_shutdown_control (self, phase, time_in_this_phase):
        if phase == 3:
            self.reject ()
            if self.writable ():
                return 1
            self.close ()
            return 0

    def isconnected (self):
        return self.connected

    def handle_timeout (self):
        self.log ("killing zombie channel %s" % ":".join (map (str, self.addr)))
        self.close ()

    def set_timeout (self, timeout):
        self.zombie_timeout = timeout

    def set_socket_timeout (self, timeout):
        self.keep_alive = timeout
        self.network_timeout = timeout

    def attend_to (self, thing):
        if not thing: return
        self.producers_attend_to.append (thing)

    def die_with (self, thing, tag):
        if not thing: return
        self.things_die_with.append ((thing, tag))

    def done_request (self):
        self.producers_attend_to = [] # all producers are finished
        self.set_timeout (self.keep_alive)

    def send (self, data):
        #print    ("SEND", repr (data), self.get_terminator ())
        self.event_time = int (time.time())
        result = asynchat.async_chat.send (self, data)
        self.server.bytes_out.inc (result)
        self.bytes_out.inc (result)
        return result

    def recv (self, buffer_size):
        self.event_time = int (time.time())
        try:
            result = asynchat.async_chat.recv (self, buffer_size)
            if not result:
                self.handle_close ()
                return b""
            #print    ("RECV", repr(result), self.get_terminator ())
            lr = len (result)
            self.server.bytes_in.inc (lr)
            self.bytes_in.inc (lr)
            return result

        except MemoryError:
            lifetime.shutdown (1, 1.0)

    def collect_incoming_data (self, data):
        #print ("collect_incoming_data", repr (data [:180]), self.current_request)
        if self.current_request:
            self.current_request.collect_incoming_data (data)
        else:
            self.in_buffer = self.in_buffer + data

    def forward_domain (self, r):
        host = r.get_header ('host', '').split (":", 1)[0]
        if host == self.server.single_domain:
            return False
        scheme, port = r.get_scheme (), self.server.port
        if (scheme == 'https' and port == 443) or (scheme == 'http' and port == 80):
            port = ""
        else:
            port = ":{}".format (port)
        newloc = '{}://{}{}{}'.format (scheme, self.server.single_domain, port, r.uri)
        r.response.set_header ("Location", newloc)
        r.response.error (301, why = "Object moved to <a href='{}'>here</a>".format (newloc))
        return True

    def found_terminator (self):
        if self.is_rejected:
            return

        if self.current_request:
            self.current_request.found_terminator()
            if DEBUG:
                self.__history.append ("REQUEST DONE")
                self.__history = self.__history [-30:]

        else:
            header = self.in_buffer
            self.in_buffer = b''
            lines = header.decode ("utf8").split('\r\n')
            while lines and not lines[0]:
                lines = lines[1:]

            if not lines:
                self.close_when_done()
                return

            request = lines[0]
            try:
                command, uri, version = http_util.crack_request (request)
            except:
                self.log_info ("channel %s-%s invalid request header" % (self.server.worker_ident, self.channel_number), "fail")
                return self.close ()

            if DEBUG:
                self.__history.append ("START REQUEST: %s/%s %s" % (command, version, uri))

            header = http_util.join_headers (lines[1:])
            r = http_request.http_request (self, request, command, uri, version, header)
            if self.server.single_domain and self.forward_domain (r):
                return
            self.set_timeout (self.network_timeout)
            self.request_counter.inc()
            self.server.total_requests.inc()

            if command is None:
                r.response.error (400)
                return

            for h in self.server.handlers:
                if h.match (r):
                    try:
                        self.current_request = r
                        h.handle_request (r)

                    except:
                        self.server.trace()
                        try: r.response.error (500)
                        except: pass

                    return

            try: r.response.error (404)
            except: pass

    def handle_abort (self):
        self.close (ignore_die_partner = True)
        self.log_info ("channel %s-%s aborted" % (self.server.worker_ident, self.channel_number), "info")

    def close (self, ignore_die_partner = False):
        if self.closed:
            return

        for closable, tag in self.things_die_with:
            if closable and hasattr (closable, "channel"):
                closable.channel = None
            self.journal (tag)
            if not ignore_die_partner:
                self.producers_attend_to.append (closable)

        if self.current_request is not None:
            self.producers_attend_to.append (self.current_request.collector)
            self.producers_attend_to.append (self.current_request.producer)
            # 1. close forcely or by error, make sure that channel is None
            # 2. by close_when_done ()
            self.current_request.channel = None
            self.current_request = None

        for closable in self.producers_attend_to:
            if closable and hasattr (closable, "close"):
                try:
                    closable.close ()
                except:
                    self.server.trace()

        self.producers_attend_to, self.things_die_with = [], []
        self.discard_buffers ()
        asynchat.async_chat.close (self)
        self.connected = False
        self.closed = True
        self.log_info ("channel %s-%s closed" % (self.server.worker_ident, self.channel_number), "info")

    def journal (self, reporter):
        self.log (
            "%s closed, client %s:%s, bytes in: %s, bytes out: %s for %d seconds " % (
                reporter,
                self.addr [0],
                self.addr [1],
                self.bytes_in,
                self.bytes_out,
                time.time () - self.creation_time
            )
        )

    def log (self, message, type = "info"):
        self.server.log (message, type)
    log_info = log

    def trace (self, id = None):
        self.server.trace (id)

    def handle_expt(self):
        self.log_info ("channel %s-%s panic" % (self.server.worker_ident, self.channel_number), "fail")
        self.close ()

    def handle_error (self):
        self.server.trace ("channel %s-%s" % (self.server.worker_ident, self.channel_number))
        self.close()


#-------------------------------------------------------------------
# server class
#-------------------------------------------------------------------
class http_server (asyncore.dispatcher):
    SERVER_IDENT = skitai.NAME
    KEEP_PRIVILEGES = False

    maintern_interval = 0
    critical_point_cpu_overload = 90.0
    critical_point_continuous = 3
    altsvc = None
    sock_type = socket.SOCK_STREAM

    def __init__ (self, ip, port, server_logger = None, request_logger = None):
        global PID

        self.handlers = []
        self.ip = ip
        self.port = port
        asyncore.dispatcher.__init__ (self)

        if ip.find (":") != -1:
            self.create_socket (socket.AF_INET6)
        else:
            self.create_socket (socket.AF_INET)

        self.set_reuse_addr ()
        try:
            self.bind ((ip, port))
        except OSError as why:
            if why.errno == 98:
                server_logger ("address already in use, cannot start server", "fatal")
            else:
                server_logger.trace ()
            sys.exit (0)

        self.worker_ident = "master"
        self.server_logger = server_logger
        self.request_logger = request_logger
        self.start_time = time.ctime(time.time())
        self.start_time_int = time.time()

        self.total_clients = counter.counter()
        self.total_requests = counter.counter()
        self.exceptions = counter.counter()
        self.bytes_out = counter.counter()
        self.bytes_in  = counter.counter()
        self.shutdown_phase = 2

        self.__last_maintern = time.time ()

        host, port = self.socket.getsockname()
        hostname = socket.gethostname()
        try:
            self.server_name = socket.gethostbyaddr (hostname)[0]
        except socket.error:
            self.server_name = hostname
        self.hash_id = md5 (self.server_name.encode ('utf8')).hexdigest() [:4]
        self.server_port = port
        self.single_domain = None

    def set_single_domain (self, domain):
        self.single_domain = domain

    def _serve (self, shutdown_phase = 2):
        self.shutdown_phase = shutdown_phase
        self.listen (os.name == "posix" and 4096 or 256)

    def serve (self, sub_server = None):
        if sub_server:
            sub_server._serve (max (1, self.shutdown_phase - 1))
        self._serve ()

    def fork (self, numworker = 1):
        global ACTIVE_WORKERS, SURVAIL, PID, EXITCODE

        if os.name == "nt":
            set_process_name (skitai.get_proc_title ())
            signal.signal(signal.SIGTERM, hTERMWORKER)

        else:
            while SURVAIL:
                try:
                    if ACTIVE_WORKERS < numworker:
                        pid = os.fork ()
                        if pid == 0:
                            if os.name != 'nt' and not self.KEEP_PRIVILEGES:
                                drop_privileges (skitai.SERVICE_USER)
                            self.worker_ident = "w%d" % len (PID)
                            set_process_name ("%s:%s" % (skitai.get_proc_title (), self.worker_ident))
                            PID = {}
                            signal.signal(signal.SIGTERM, hTERMWORKER)
                            signal.signal(signal.SIGQUIT, hQUITWORKER)
                            break

                        else:
                            set_process_name ("%s:m" % skitai.get_proc_title ())
                            if not PID:
                                signal.signal(signal.SIGHUP, hHUPMASTER)
                                signal.signal(signal.SIGTERM, hTERMMASTER)
                                signal.signal(signal.SIGINT, hTERMMASTER)
                                signal.signal(signal.SIGQUIT, hQUITMASTER)
                                signal.signal (signal.SIGCHLD, hCHLD)

                            ps = psutil.Process (pid)
                            ps.x_overloads = 0
                            PID [pid] = ps
                            ACTIVE_WORKERS += 1
                            #print ('-----', PID, ACTIVE_WORKERS)

                    now = time.time ()
                    if self.maintern_interval and now - self.__last_maintern > self.maintern_interval:
                        self.maintern (now)
                    time.sleep (1)

                except KeyboardInterrupt:
                    EXITCODE = 0
                    DO_SHUTDOWN (signal.SIGTERM)
                    break

            if self.worker_ident == "master":
                return EXITCODE

        self.log_info ('%s%s(%s) started on %s:%d' % (
            self.SERVER_IDENT, hasattr (self, 'ctx') and ' SSL' or ' ', self.worker_ident, self.server_name, self.port)
        )
        if self.altsvc:
            self.log_info ('%s QUIC(%s) started on %s:%d' % (
                self.SERVER_IDENT, self.worker_ident, self.server_name, self.altsvc.port)
            )

    usages = []
    cpu_stat = []
    def maintern (self, now):
        global PID, CPUs

        self.__last_maintern = now
        usages = []
        for ps in PID.values ():
            if ps is None:
                continue

            try:
                usage = ps.cpu_percent ()
            except (psutil.NoSuchProcess, AttributeError):
                # process changed, maybe next time
                usage = 0.0
            usages.append ((ps, usage))

        if not usages:
            # restarting?
            return

        self.usages.append (sum ([x [1] for x in usages]) / len (usages))
        self.usages = self.usages [-45:]
        self.cpu_stat = []
        max_usage = 0
        for c in (3, 15, 45):
            # 1m, 5m, 15
            array = self.usages [-c:]
            self.cpu_stat.append ((max (array), int (sum (array) / c)))

        if self.cpu_stat [0][0] > 10:
            # 60% for 1m
            self.log ("CPU percent: " + ", ".join (["%d/%d" % unit for unit in self.cpu_stat]))

        for ps, usage in usages:
            if usage < self.critical_point_cpu_overload:
                ps.x_overloads = 0 #reset count
                continue

            ps.x_overloads += 1
            if ps.x_overloads > self.critical_point_continuous:
                self.log ("process %d is overloading, try to kill..." % ps.pid, 'fatal')
                sig = ps.x_overloads > (self.critical_point_continuous + 2) and signal.SIGKILL or signal.SIGTERM
                try:
                    os.kill (ps.pid, sig)
                except OSError:
                    pass

    def create_socket(self, family):
        if hasattr (socket, "_no_timeoutsocket"):
            sock_class = socket._no_timeoutsocket
        else:
            sock_class = socket.socket

        self.family_and_type = family, self.sock_type
        sock = sock_class (family, self.sock_type)
        sock.setblocking(0)
        self.set_socket(sock)

    def clean_shutdown_control (self, phase, time_in_this_phase):
        if phase == self.shutdown_phase:
            self.log_info ('shutting down web server: %s' % self.server_name)
            self.close ()

    def close (self):
        asyncore.dispatcher.close (self)
        for h in self.handlers:
            if hasattr (h, "close"):
                h.close ()

    def writable (self):
        return 0

    def install_handler (self, handler, back = 1):
        if back:
            self.handlers.append (handler)
        else:
            self.handlers.insert (0, handler)

    def remove_handler (self, handler):
        self.handlers.remove (handler)

    def log (self, message, type = "info"):
        if self.server_logger:
            self.server_logger.log (message, type)
        else:
            sys.stdout.write ('log: [%s] %s\n' % (type,str (message)))

    def log_request (self, message):
        if self.request_logger:
            self.request_logger.log (message, "")
        else:
            sys.stdout.write ('%s\n' % message)

    def log_info(self, message, type='info'):
        self.log (message, type)

    def trace (self, id = None):
        self.exceptions.inc()
        if self.server_logger:
            self.server_logger.trace (id)
        else:
            asyncore.dispatcher.handle_error (self)

    def handle_read (self):
        pass

    def readable (self):
        return self.accepting

    def handle_error (self):
        self.trace()

    def handle_connect (self):
        pass

    def handle_accept (self):
        self.total_clients.inc()
        try:
            conn, addr = self.accept()
        except socket.error:
            self.log_info ('server accept() threw an exception', 'warn')
            return
        except TypeError:
            if os.name == "nt":
                self.log_info ('server accept() threw EWOULDBLOCK', 'warn')
            return
        #self.log_info ('client %s:%d accepted by %s' % (addr [0], addr [1], self.worker_ident))
        http_channel (self, conn, addr)

    def handle_expt (self):
        self.log_info ('socket panic', 'warning')

    def handle_close (self):
        self.log_info('server shutdown', 'warning')
        self.close()

    def status (self):
        global PID

        return     {
            "child_pids": list (PID.keys ()),
            "ident": "%s for %s" % (self.worker_ident, self.SERVER_IDENT),
            'server_name': self.server_name,
            "start_time": self.start_time,
            'hash_id': self.hash_id,
            "port": self.port,
            "total_clients": self.total_clients.as_long(),
            "total_request": self.total_requests.as_long(),
            "total_exceptions": self.exceptions.as_long(),
            "bytes_out": self.bytes_out.as_long(),
            "bytes_in": self.bytes_in.as_long(),
            "cpu_percent":    self.cpu_stat
        }

def hCHLD (signum, frame):
    global ACTIVE_WORKERS, PID

    ACTIVE_WORKERS -= 1
    try:
        pid, status = os.waitpid (0, 0)
    except ChildProcessError:
        pass
    else:
        PID [pid]    = None

def hTERMWORKER (signum, frame):
    lifetime.shutdown (0, 1.0)

def hQUITWORKER (signum, frame):
    lifetime.shutdown (0, 30.0)

def DO_SHUTDOWN (sig):
    global PID

    signal.signal (signal.SIGCHLD, signal.SIG_IGN)
    for pid in PID:
        if PID [pid] is None:
            continue
        try: os.kill (pid, sig)
        except OSError: pass

def hTERMMASTER (signum, frame):
    global EXITCODE, SURVAIL
    SURVAIL = False
    EXITCODE = 0
    DO_SHUTDOWN (signal.SIGTERM)

def hQUITMASTER (signum, frame):
    global EXITCODE
    EXITCODE = 0
    DO_SHUTDOWN (signal.SIGQUIT)

def hHUPMASTER (signum, frame):
    global EXITCODE
    EXITCODE = 3
    DO_SHUTDOWN (signal.SIGTERM)

def configure (name, network_timeout = 0, keep_alive = 0):
    from . import https_server
    http_server.SERVER_IDENT = name
    http_channel.keep_alive = https_server.https_channel.keep_alive = not keep_alive and 2 or keep_alive
    http_channel.network_timeout = https_server.https_channel.network_timeout = not network_timeout and 30 or network_timeout

    try:
        from . import http3_server
    except ImportError:
        pass
    else:
        http3_server.http3_channel.keep_alive = not keep_alive and 2 or keep_alive
        http3_server.http3_channel.network_timeout = not network_timeout and 30 or network_timeout


if __name__ == "__main__":
    pass



import zlib
import time
import os
import sys
from aquests.protocols.http import http_date, http_util
from rs4.reraise import reraise
from rs4 import producers, compressors
from aquests.protocols.http import respcodes
from ..wastuff import selective_logger
from ..utility import catch
import skitai
from rs4 import asyncore
import json
from skitai import exceptions
from ..wastuff.api import API, catch

try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin


UNCOMPRESS_MAX = 2048
ONETIME_COMPRESS_MAX = 1048576

# Default error message
DEFAULT_ERROR_MESSAGE = """<!DOCTYPE html>
<head>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8">
<title>%(code)d %(message)s</title>
<style type="text/css"><!-- *{font-family:sans-serif, verdana; -webkit-font-smoothing: antialiased !important;}body{margin:0;padding:0;font-size:14px;color:#1e1e1e;font-family:verdana,sans-serif;text-align:center;} #titles{padding:16px;}#titles h1{color: #000000;} #content{padding:16px 16px 30px 16px;} #debug h2 {font-size: 18px; font-weight: 600;} #debug h3{font-size: 18px; color: #7A297B;} #debug p,b,h4,li {font-size:14px;}#debug h4{color: #363636;} #debug li{margin-bottom: 6px;} #debug .f {color:#731D75; font-weight: normal;} #debug .n {color:#003366;font-weight:bold; text-align: right;} #debug{text-align: left; background-color: #FFF; margin-bottom: 32px; padding:8px 16px 8px 16px; border-radius: 3px;} #debug li,i{font-weight:normal;}#footer {font-size:12px;padding-left:10px;}#software {border-top: solid 1px #cccccc;} #software div {padding-top: 10px; font-size: 12px;} --></style>
</head>
<body>
<div id="titles"><h1>%(code)d %(message)s</h1></div>
<div id="content">
<div id="detail">%(detail)s</div>
<div id="traceback">
    <div id="%(mode)s"><p>%(traceback)s</p></div>
</div>
<div id="software"><div>%(software)s</div></div>
</div>
</body>
</html>"""


class http_response:
    log_or_not = selective_logger.SelectiveLogger ()
    reply_code = 200
    reply_message = "OK"
    _is_async_streaming = False

    def __init__ (self, request):
        self.request = request
        self.reply_headers = [
            ('Server', skitai.NAME),
            ('Date', http_date.build_http_date (time.time()))
        ]
        altsvc = self.request.channel.server.altsvc
        if altsvc:
            self.reply_headers.append (("Alt-Svc", altsvc.ALTSVC_HEADER))
        self.outgoing = producers.fifo ()
        self._is_done = False
        self.stime = time.time ()
        self.htime = 0
        self.content_type = None
        self.current_app = None
        self.current_env = None

    def is_async_streaming (self):
        return self._is_async_streaming

    def set_streaming (self):
        self._is_async_streaming = True

    def is_responsable (self):
        return not self._is_done

    def is_done (self):
        return self._is_done

    def __len__ (self):
        return len (self.outgoing)

    def __setitem__ (self, key, value):
        self.set (key, value)

    def __getitem__ (self, key):
        return self.get (key)

    def __delitem__ (self, k):
        self.delete (k)

    def has_key (self, key):
        key = key.lower ()
        return key in [x [0].lower () for x in self.reply_headers]

    def set (self, key, value):
        self.reply_headers.append ((key, value))

    def get (self, key):
        key = key.lower ()
        for k, v in self.reply_headers:
            if k.lower () == key:
                return v

    def delete (self, key):
        index = 0
        found = 0
        key = key.lower ()
        for hk, hv in self.reply_headers:
            if key == hk.lower ():
                found = 1
                break
            index += 1

        if found:
            del self.reply_headers [index]
            self.delete (key)

    def update (self, key, value):
        self.delete (key)
        self.set (key, value)

    def append_header (self, key, value):
        val = self.get (key)
        if not val:
            self.set (key, value)
        else:
            self.set (key, val + ", " + value)

    set_header = set
    get_header = get
    del_header = delete

    def set_cache (self, max_age = 0):
        if self.request.version == "1.0":
            self.set_header ('Expires', http_date.build_http_date (time.time() + max_age))
        else:
            self.set_header ('Cache-Control', "max-age=%d" % max_age)

    def set_headers (self, headers):
        for k, v in headers:
            self.set (k, v)

    def append_headers (self, headers):
        for k, v in headers:
            self.set (k, v)

    def get_hedaers (self):
        return self.reply_headers

    def set_reply (self, status):
        # for atila.Atila App
        self.reply_code, self.reply_message = self.parse_ststus (status)
    set_status = set_reply

    def get_status (self):
        return "%d %s" % self.get_reply ()

    def build_reply_header (self, with_header = 1):
        h = [self.response (self.reply_code, self.reply_message)]
        if with_header:
            h.extend (['%s: %s' % x for x in self.reply_headers])
        h = '\r\n'.join (h) + '\r\n\r\n'
        return h

    def get_status_msg (self, code):
        return respcodes.get (code, "Undefined Error")

    def response (self, code, status):
        return 'HTTP/%s %d %s' % (self.request.version, code, status)

    def parse_ststus (self, status):
        try:
            code, status = status.split (" ", 1)
            code = int (code)
        except:
            raise AssertionError ("Can't understand given status code")
        return code, status

    def get_reply (self):
        # for atila.Atila App
        return self.reply_code, self.reply_message

    def instant (self, status = "", headers = None):
        # instance messaging
        if self.request.channel is None:
            return
        code, msg = self.parse_ststus (status)
        reply = [self.response (code, msg)]
        if headers:
            for header in headers:
                reply.append ("%s: %s" % header)
        self.request.channel.push (("\r\n".join (reply) + "\r\n\r\n").encode ("utf8"))

    # for WSGI App -------------------------------------------------------
    def start_response (self, status, headers = None, exc_info = None):
        if not self.is_responsable ():
            if exc_info:
                try:
                    reraise (*exc_info)
                finally:
                    exc_info = None
            else:
                raise AssertionError ("Response already sent!")
            return

        code, status = self.parse_ststus (status)
        self.start (code, status, headers)

        if exc_info:
            # expect plain/text, send exception info to developers
            content = catch (0, exc_info)
            self.push (content)

        return self.push #by WSGI Spec.

    # Internal Response Methods ------------------------------------------
    def abort (self, code, status = "", why = ""):
        self.request.channel.reject ()
        self.error (code, status, why, force_close = True)

    def start (self, code, status = "", headers = None):
        if not self.is_responsable ():
            return
        self.reply_code = code
        if status:
            self.reply_message = status
        else:
            self.reply_message = self.get_status_msg (code)
        if headers:
            if isinstance (headers, dict):
                headers = headers.items ()
            for k, v in headers:
                self.set (k, v)
    reply = start

    def catch (self, exc_info = None):
        if self.request.get_header ('accept').find ("application/json") != -1:
            return catch (2, exc_info)
        return catch (1, exc_info)

    def build_error_template (self, why = '', errno = 0, was = None):
        global DEFAULT_ERROR_MESSAGE

        exc_info = None
        if type (why) is tuple: # sys.exc_info ()
            if self.current_app and not self.current_app.debug:
                why = None
            exc_info, why  = why, ''

        is_html_response = self.request.get_header ('accept', '').find ("text/html") != -1
        error = {
            'code': self.reply_code,
            'errno': errno,
            'message': self.reply_message,
            'detail': why,
            'mode': exc_info and 'debug' or 'normal',
            'time': http_date.build_http_date (time.time ()),
            'url': urljoin ("%s://%s/" % (self.request.get_scheme (), self.request.get_header ("host")), self.request.uri),
            'software': skitai.NAME,
            'debug': None,
            "traceback": exc_info and catch (is_html_response and 1 or 2, exc_info) or None,
        }
        if self.current_app and hasattr (self.current_app, 'render_error'):
            content = None
            try:
                content = self.current_app.render_error (error, was)
            except:
                self.request.logger.trace ()
                if self.current_app.debug:
                    if is_html_response:
                        error ["traceback"] += "<h2 style='padding-top: 40px;'>Exception Occured During Building Error Template</h2>" + catch (1)
                    else:
                        error ["traceback"] += ["Exception Occured During Building Error Template"] + catch (2)
            if content:
                return content

        if is_html_response:
            self.update ('content-type', 'text/html')
            return DEFAULT_ERROR_MESSAGE % {k: v or '' for k, v in error.items ()}
        else:
            return self.fault (
                error ["message"].lower (), errno,
                None,
                error ["detail"],
                exc_info = error ["traceback"]
            )

    def error (self, code, status = "", why = "", force_close = False, push_only = False):
        if not self.is_responsable (): return
        self.outgoing.clear ()
        self.reply_code = code
        if status: self.reply_message = status
        else:    self.reply_message = self.get_status_msg (code)

        body = self.build_error_template (why).encode ("utf8")
        self.update ('Content-Length', len(body))
        self.update ('Cache-Control', 'max-age=0')
        self.push (body)
        if not push_only:
            self.done (force_close)

    # Send Response -------------------------------------------------------
    def die_with (self, thing):
        if self.request.channel:
            self.request.channel.attend_to (thing)

    def push_promise (self, *args, **kargs):
        # ignore in version 1.x
        pass
    hint_promise = push_promise

    def push_and_done (self, thing, *args, **kargs):
        self.push (thing)
        self.done (*args, **kargs)

    def push (self, thing):
        if not self.is_responsable ():
            return
        if type(thing) is bytes:
            self.outgoing.push (producers.simple_producer (thing))
        else:
            self.outgoing.push (thing)

    def __call__ (self, status = "200 OK", body = None, headers = None, exc_info = None):
        self.start_response (status, headers)
        if body is None:
            return self.build_error_template (exc_info)
        return body

    def done (self, force_close = False, upgrade_to = None, with_header = 1):
        self.content_type = self.get ('content-type')

        if not self.is_responsable (): return
        self._is_done = True
        if self.request.channel is None: return

        self.htime = (time.time () - self.stime) * 1000
        self.stime = time.time () #for delivery time

        # compress payload and globbing production
        do_optimize = True
        if upgrade_to or self.is_async_streaming ():
            do_optimize = False

        connection = http_util.get_header (http_util.CONNECTION, self.request.header).lower()
        close_it = False
        way_to_compress = ""
        wrap_in_chunking = False

        if force_close:
            close_it = True
            if self.request.version == '1.1':
                self.update ('Connection', 'close')
            else:
                self.delete ('Connection')

        else:
            if self.request.version == '1.0':
                if connection == 'keep-alive':
                    if not self.has_key ('content-length'):
                        close_it = True
                        self.update ('Connection', 'close')
                    else:
                        self.update ('Connection', 'keep-alive')
                else:
                    close_it = True

            elif self.request.version == '1.1':
                if connection == 'close':
                    close_it = True
                    self.update ('Connection', 'close')
                if not self.has_key ('transfer-encoding') and not self.has_key ('content-length') and self.has_key ('content-type'):
                    wrap_in_chunking = True

            else:
                # unknown close
                self.update ('Connection', 'close')
                close_it = True

        if len (self.outgoing) == 0:
            self.update ('Content-Length', "0")
            self.delete ('transfer-encoding')
            self.delete ('content-type')
            outgoing_producer = producers.simple_producer (self.build_reply_header(with_header).encode ("utf8"))
            do_optimize = False

        elif len (self.outgoing) == 1 and hasattr (self.outgoing.first (), "ready"):
            outgoing_producer = producers.composite_producer (self.outgoing)
            if wrap_in_chunking:
                self.update ('Transfer-Encoding', 'chunked')
                outgoing_producer = producers.chunked_producer (outgoing_producer)
            outgoing_header = producers.simple_producer (self.build_reply_header(with_header).encode ("utf8"))
            self.request.channel.push_with_producer (outgoing_header)
            do_optimize = False

        elif do_optimize and not self.has_key ('Content-Encoding'):
            maybe_compress = self.request.get_header ("Accept-Encoding")
            if maybe_compress:
                cl = self.has_key ("content-length") and int (self.get ("Content-Length")) or -1
                if cl == -1:
                    cl = self.outgoing.get_estimate_content_length ()

                if 0 < cl  <= UNCOMPRESS_MAX:
                    maybe_compress = ""
                elif not wrap_in_chunking and cl > ONETIME_COMPRESS_MAX:
                    # too big for memory, do not compress
                    maybe_compress = ""

            if maybe_compress:
                content_type = self.get ("Content-Type")
                if content_type and (content_type.startswith ("text/") or content_type.startswith ("application/json")):
                    accept_encoding = [x.strip () for x in maybe_compress.split (",")]
                    if "gzip" in accept_encoding:
                        way_to_compress = "gzip"
                    elif "deflate" in accept_encoding:
                        way_to_compress = "deflate"

            if way_to_compress:
                if self.has_key ('Content-Length'):
                    self.delete ("content-length") # rebuild
                self.update ('Content-Encoding', way_to_compress)

            if wrap_in_chunking:
                outgoing_producer = producers.composite_producer (self.outgoing)
                self.delete ('content-length')
                self.update ('Transfer-Encoding', 'chunked')
                if way_to_compress:
                    if way_to_compress == "gzip":
                        compressing_producer = producers.gzipped_producer
                    else: # deflate
                        compressing_producer = producers.compressed_producer
                    outgoing_producer = compressing_producer (outgoing_producer)
                outgoing_producer = producers.chunked_producer (outgoing_producer)
                outgoing_header = producers.simple_producer (self.build_reply_header(with_header).encode ("utf8"))

            else:
                self.delete ('transfer-encoding')
                if way_to_compress:
                    if way_to_compress == "gzip":
                        compressor = compressors.GZipCompressor ()
                    else: # deflate
                        compressor = zlib.compressobj (6, zlib.DEFLATED)
                    cdata = b""
                    has_producer = 1
                    while 1:
                        has_producer, producer = self.outgoing.pop ()
                        if not has_producer: break
                        while 1:
                            data = producer.more ()
                            if not data:
                                break
                            cdata += compressor.compress (data)
                    cdata += compressor.flush ()
                    self.update ("Content-Length", len (cdata))
                    outgoing_producer = producers.simple_producer (cdata)
                else:
                    outgoing_producer = producers.composite_producer (self.outgoing)

                outgoing_header = producers.simple_producer (self.build_reply_header(with_header).encode ("utf8"))

            outgoing_producer = producers.composite_producer (
                producers.fifo([outgoing_header, outgoing_producer])
            )

        outgoing_producer = self.log_or_not (self.request.uri, outgoing_producer, self.log)
        if do_optimize:
            outgoing_producer = producers.globbing_producer (outgoing_producer)

        # IMP: second testing after push_with_producer()->init_send ()
        if self.request.channel is None: return

        if upgrade_to:
            request, terminator = upgrade_to
            self.request.channel.current_request = request
            self.request.channel.set_terminator (terminator)
        else:
            # preapre to receice new request for channel
            self.request.channel.current_request = None
            self.request.channel.set_terminator (b"\r\n\r\n")

        # proxy collector and producer is related to asynconnect
        # and relay data with channel
        # then if request is suddenly stopped, make sure close them
        self.die_with (self.request.collector)
        self.die_with (self.request.producer)

        logger = self.request.logger #IMP: for  disconnect with request
        try:
            if outgoing_producer:
                self.request.channel.push_with_producer (outgoing_producer)
            if close_it:
                self.request.channel.close_when_done ()
        except:
            logger.trace ()

    def maybe_log (self, bytes):
        if self.log_or_not.loggable (self.request.uri):
            self.log (bytes)

    def log (self, bytes):
        server = self.request.channel.server
        referer = self.request.get_header ('referer')
        real_ip = self.request.get_real_ip ()
        worker = server.worker_ident [0] == "m" and "M" or ("W" + server.worker_ident [1:])
        host = self.request.host
        server.log_request (
            '%s %s %s %s %d %d %d %s %d %s %s %s %s %s %s %s %s %d %d'
            % (
            self.request.channel.addr [0],
            self.request.is_promise () and "PUSH" or self.request.method,
            self.request.uri,
            self.request.version,
            self.request.rbytes, # recv body size

            self.reply_code,
            bytes, # send body size
            worker,
            self.htime, # due time to request handling

            self.request.user_agent and '"' + self.request.user_agent + '"' or "-",
            host and host [0].isalpha () and host or "-",
            referer and '"' + referer + '"' or "-",
            real_ip != self.request.channel.addr [0] and real_ip or '-',

            # debugging
            self.request.gtxid or "-",
            self.request.ltxid or "-",
            self.request.user and '"' + str (self.request.user) + '"' or "-",
            self.request.token or "-",

            len (asyncore.socket_map),
            (time.time () - self.stime) * 1000, # due time to sending data
            )
        )
        # clearing resources, back refs
        self.request.response_finished ()

    # Sugar syntaxes ----------------------------------------------------------------
    def set_timeout (self, timeout):
        self.request.channel.set_timeout (timeout)

    def adaptive_error (self, status, message, code, more_info):
        ac = self.request.get_header ('accept', '')
        if ac.find ("text/html") != -1:
            return self.with_explain (status, "{} (code: {}): {}".format (message, code, more_info))
        return self.Fault (status, message, code, None, more_info)

    def fault (self, message = "", code = 0, debug = None, more_info = None, exc_info = None, traceback = False):
        api = self.API ()
        if not code:
            code = int (self.reply_code) * 100 + (traceback and 90 or 0)
        if traceback:
            api.traceback (message, code, debug or "see traceback", more_info)
        else:
            api.error (message, code, debug, more_info, exc_info)
        return api
    eapi = fault # will be derecating

    def throw (self, status, why = ""):
        raise exceptions.HTTPError (status, why)

    def with_explain (self, status = "200 OK", why = "", errcode = None, headers = None):
        self.start_response (status, headers)
        return self.build_error_template (why, errcode)

    def API (self, __status__ = None, __data_dict__ = None, **kargs):
        if isinstance (__status__, str):
            self.set_status (__status__)
        elif isinstance (__status__, dict):
            __data_dict__ = __status__
        elif __status__ is not None:
            raise TypeError ("first arg for was.API should be status string or dict data only")
        if __data_dict__ and kargs:
            __data_dict__.update (kargs)
        api = API (self.request, __data_dict__ or kargs)
        self.update ("Content-Type", api.get_content_type ())
        return api
    api = API

    def Fault (self, status = "200 OK", *args, **kargs):
        self.set_status (status)
        r = self.fault (*args, **kargs)
        return self (status, r)
    for_api = Fault

    def File (self, path, mimetype = 'application/octet-stream', filename = None):
        self.set_header ('Content-Type',  mimetype)
        if filename:
            self.set_header ('Content-Disposition', 'attachment; filename="{}"'.format (filename))

        if isinstance (path, str):
            self.set_header ('Content-Length', str (os.path.getsize (path)))
            fp = open (path, "rb")
        else:
            fp = path
            fp.seek (0)
        return producers.file_producer (fp)

    file = File



from skitai import counter
from rs4 import compressors
from hashlib import md5
from rs4 import pathtool
import os
import time
import threading
import glob
import threading
from functools import reduce

class CacheInfo:
	def __init__ (self):
		self.count = 0
		self.allocated = 0.0
		self.accumulated_caches = 0
		self.accumulated_expires = 0

	def inc (self, delta):
		self.count += 1
		self.allocated += delta
		self.accumulated_caches += 1

	def dec (self, delta):
		self.count -= 1
		self.allocated -= delta
		self.accumulated_expires += 1

	def get (self):
		return (
			self.count,
			self.allocated,
			self.accumulated_caches,
			self.accumulated_expires
		)


class CacheFileSystem:
	maintern_interval = 600

	def __init__ (self, path = None, memmax = 64, diskmax = 0):
		self.path = path
		if not self.path:
			diskmax = 0
		self.max_memory = memmax * 1024 * 1024
		self.max_disk = diskmax * 1024 *1024
		self.files = {}
		self.current_memory = CacheInfo ()
		self.current_disk = CacheInfo ()
		self.maintern_times = {}
		self.lock = threading.RLock ()
		self.mainterns = counter.counter ()
		self.numhits = counter.counter ()
		self.numfails = counter.counter ()
		if self.max_disk:
			self.check_dir ()

	def status (self):
		with self.lock:
			r = {
				"numhits": self.numhits.as_long (),
				"numfails": self.numfails.as_long (),
				'mainterned':self.mainterns.as_long (),
				"memory": self.current_memory.get (),
				"disk": self.current_disk.get ()
			}
		return r

	def check_dir (self):
		for h in list(range(48,58)) + list(range(97,103)):
			for i in list(range(48,58)) + list(range(97,103)):
				for j in list(range(48,58)) + list(range(97,103)):
					initial = "0" + chr (h) + "/" + chr (i) + chr (j)
					pathtool.mkdir (os.path.join (self.path, initial))
					self.load_cache (initial)

	def load_cache (self, initial, force_remove = False):
		c = 0
		current_time = time.time ()
		for path in glob.glob (os.path.join (self.path, initial, "*")):
			created = os.stat (path).st_mtime

			try: self.files [initial]
			except KeyError: 	self.files [initial] = {}

			fn = os.path.split (path)[-1]
			f = open (path, "rb")
			max_age = int (f.read (12).strip ())
			compressed = int (f.read (1).strip ())
			f.close ()

			if created < current_time - max_age:
				try: os.remove (path)
				except (OSError, IOError): pass
				continue

			size = os.path.getsize (path)
			self.files [initial][fn] = (created, 1, size, compressed, max_age)
			with self.lock:
				self.current_disk.inc (size)

	def maintern (self, initial, force_remove = False):
		current_time = time.time ()
		deletables = []
		for fn in self.files [initial]:
			cached = self.files [initial][fn]
			valid_time = current_time - cached [4] # max-age
			if cached [0] < valid_time or force_remove:
				deletables.append ((fn, cached [1], cached [2]))

		for fn, memhit, usage in deletables:
			if memhit == 1:
				with self.lock:
					self.current_memory.dec (usage)
				del self.files [initial][fn]
			elif self.path:
				try: os.remove (os.path.join (self.path, initial, fn))
				except (OSError, IOError): pass
				else:
					with self.lock:
						self.current_disk.dec (usage)
					del self.files [initial][fn]

		self.maintern_times [initial] = current_time
		with self.lock:
			self.mainterns.inc ()

	def truncate (self):
		for initial in self.files.keys ():
			self.maintern (initial, True)

	def getpath (self, key):
		key = key.encode ("utf8")
		file = md5 (key).hexdigest ()
		initial = "0" + file [0] + "/" + file [1:3]
		if self.path:
			path = os.path.join (self.path, initial, file), initial, file
		else:
			path = None
		return path, initial, file

	def is_cachable (self, method, cache_control, has_cookie, has_auth, pragma):
		if method != "GET":
			return False
		if pragma == "no-cache":
			return False
		if cache_control:
			cache_control = list (map (lambda x: x.strip (), cache_control.split (",")))
			if "no-cache" in cache_control or "max-age=0" in cache_control:
				return False
			if has_auth and "public" not in cache_control:
				cachable = False
			if has_cookie and "public" not in cache_control:
				cachable = False
		return True

	def get (self, key, undecompressible = 0):
		path, initial, fn = self.getpath (key)

		try:
			cached = self.files [initial][fn]
		except KeyError:
			with self.lock:
				self.numfails.inc ()
			return None, None, None, None, None

		current_time = int (time.time ())
		memhit = cached [1]
		max_age = cached [4]

		if cached [0] < current_time - max_age:
			if memhit == 1:
				del self.files [initial][fn]
				with self.lock:
					self.current_memory.dec (cached [2])
			elif path:
				try:
					os.remove (path)
				except (OSError, IOError):
					pass
				else:
					del self.files [initial][fn]
					with self.lock:
						self.current_disk.dec (cached [2])
			return None, None, None, None, None

		with self.lock:
			self.numhits.inc ()

		if memhit == 1:
			content_type, content = cached [-2:]

		else:
			f = open (path, "rb")
			f.read (13) # abandon max_age, compressed
			content_type = f.read (64).strip ()
			content = f.read ()
			f.close ()

		compressed = cached [3]
		if compressed and not undecompressible:
			decompressor = compressors.GZipDecompressor ()
			content = decompressor.decompress (content) + decompressor.flush ()
			compressed = 0

		return memhit, compressed, max_age, content_type, content

	def save (self, key, content_type, content, max_age, compressed = 0):
		if self.max_memory == 0 or not self.max_disk == 0:
			return

		usage = len (content)
		if usage > 10000000:
			return

		if len (str (max_age)) > 12 or len (content_type) > 64:
			return

		# check memory status
		with self.lock:
			current_memory = self.current_memory.get ()[1]
			current_disk = self.current_disk.get ()[1]
		if current_memory > self.max_memory and current_disk > self.max_disk:
			# there's no memory/disk room for cache
			return self.maintern (initial)

		current_time = int (time.time ())
		path, initial, fn = self.getpath (key)
		try: self.files [initial]
		except KeyError: self.files [initial] = {}
		else:
			with self.lock:
				last_maintern = self.maintern_times.get (initial, 0.)
			if last_maintern == 0:
				self.maintern_times [initial] = current_time
			elif current_time - last_maintern > self.maintern_interval:
				self.maintern (initial)

		# already have valid cache
		cached = self.files [initial].get (fn)
		if cached: # already have
			return

		if not compressed and content_type.startswith ("text/"):
			compressor = compressors.GZipCompressor ()
			content = compressor.compress (content) + compressor.flush ()
			compressed = 1
			usage = len (content)

		if current_memory <= self.max_memory:
			usage *= 1.5
			self.files [initial][fn] = (current_time, 1, usage, compressed, max_age, content_type, content)
			with self.lock:
				self.current_memory.inc (usage)
			return

		if self.max_disk:
			f = open (path, "wb")
			if not content_type:
				content_type = b""
			f.write (("%12s%d%64s" % (max_age, compressed, content_type)).encode ("utf8"))
			f.write (content)
			f.close ()
			self.files [initial][fn] = (current_time, -1, usage, compressed, max_age)
			with self.lock:
				self.current_disk.inc (usage)


if __name__ == "__main__":
	f = ReverseProxy ("g:\\ttt")
	f.save ("a", "b", "asdasda")
	print(f.get ("a", "b"))
	print(f.status ())



import json
import sys
from xmlrpc import client
from ..utility import catch
import copy
from datetime import timezone, datetime, date

TZ_LOCAL = datetime.now (timezone.utc).astimezone().tzinfo
TZ_UTC = timezone.utc

class StringDateTimeEncoder (json.JSONEncoder):
	def _to_utc (self, obj):
		try:
			return obj.astimezone (TZ_UTC)
		except ValueError:
			return obj.replace (tzinfo = TZ_LOCAL).astimezone (TZ_UTC)

	def default (self, obj):
		if isinstance (obj, date):
			return str (obj)
		return json.JSONEncoder.default (self, obj)

class DigitTimeEncoder (StringDateTimeEncoder):
	def default (self, obj):
		if isinstance (obj, date):
			return self._to_utc (obj).strftime ('%Y%m%d%H%M%S')
		return json.JSONEncoder.default (self, obj)

class ISODateTimeEncoder (StringDateTimeEncoder):
	def default (self, obj):
		if isinstance (obj, date):
			return self._to_utc (obj).isoformat ()
		return json.JSONEncoder.default (self, obj)

class UNIXEpochDateTimeEncoder (StringDateTimeEncoder):
	def default (self, obj):
		if isinstance (obj, date):
			return self._to_utc (obj).timestamp ()
		return json.JSONEncoder.default (self, obj)

class ISODateTimeWithOffsetEncoder (StringDateTimeEncoder):
	def default (self, obj):
		if isinstance (obj, date):
			return self._to_utc (obj).strftime ('%Y-%m-%d %H:%M:%S+00')
		return json.JSONEncoder.default (self, obj)


class API:
	ENCODER_MAP = {
		'str': StringDateTimeEncoder,
		'iso': ISODateTimeEncoder,
		'unixepoch': UNIXEpochDateTimeEncoder,
		'digit': DigitTimeEncoder,
		'utcoffset': ISODateTimeWithOffsetEncoder
	}

	@classmethod
	def add_custom_encoder (cls, name, encoder):
		cls.ENCODER_MAP [name] = encoder

	def __init__ (self, request, data = None):
		self.request = request # for response typing
		self.data = data or {}
		self.content_type = self.set_content_type ()
		self.data_encoder_class = None

	def set_json_encoder (self, encoder):
		if not encoder:
			return
		if isinstance (encoder, str):
			self.data_encoder_class = self.ENCODER_MAP [encoder]
			return
		self.data_encoder_class = encoder

	def __setitem__ (self, k, v):
		self.set (k, v)

	def __getitem__ (self, k, v = None):
		return self.get (k, v)

	def set (self, k, v):
		self.data [k] = v

	def get (self, k, v = None):
		return self.data.get (k, v)

	def get_content_type (self):
		return self.content_type

	def set_content_type (self):
		content_type = self.request.get_header ("accept", 'application/json')
		if not content_type.startswith ("text/xml"):
			content_type = 'application/json'
		return content_type

	def encode (self, charset):
		return self.to_string ().encode (charset)

	def __str__ (self):
		return self.to_string ()

	def to_string (self):
		if self.content_type.startswith ("text/xml"):
			return client.dumps ((self.data,))
		return json.dumps (self.data, cls = self.data_encoder_class or self.ENCODER_MAP ['utcoffset'], ensure_ascii = False, indent = 2)

	def traceback (self, message = 'exception occured', code = 20000, debug = 'see traceback', more_info = None):
		self.error (message, code, debug, more_info, sys.exc_info ())

	def error (self, message, code = 20000, debug = None, more_info = None, exc_info = None):
		self.data = {}
		self.data ['code'] = code
		self.data ['message'] = message
		if debug:
			self.data ['debug'] = debug
		if more_info:
			self.data ['more_info'] = more_info
		if exc_info:
			self.data ["traceback"] = type (exc_info) is tuple and catch (2, exc_info) or exc_info

	def set_spec (self, app):
		resource_id = self.request.routable ["func_id"]
		routable = copy.deepcopy (self.request.routable)
		del routable ["func_id"]
		routable ["methods"] = list (routable ["methods"])
		props = {"current_request": {}}
		props ["current_request"]["method"] = self.request.method
		props ["current_request"]["version"] = self.request.version
		props ["current_request"]["uri"] = self.request.uri
		props ['routeopt'] = routable
		props ["parameter_requirements"] = app.get_parameter_requirements (resource_id)
		props ['auth_requirements'] = app.get_auth_flags (resource_id)
		props ["doc"] = self.request.routed.__doc__
		props ["id"] = resource_id
		self.data ["__spec__"] = props

		#import pprint
		#pprint.pprint (self.data ["__spec__"])



#!/usr/bin/python3
# 2015. 11. 27 by Hans Roh hansroh@gmail.com

import sys, os, getopt
from skitai import lifetime
from rs4 import pathtool, logger, confparse
from aquests.athreads import select_trigger
from rs4.psutil import daemon as demonizer, service, daemon_class
from aquests.protocols.smtp import async_smtp, composer
import signal
import time
import glob

DEFAULT = """[smtpda]
# SMTP Delivery Agent
max-retry = 3
keep-days = 3
smtp-server =
ssl = no
user =
password =
"""


def hTERM (signum, frame):
	lifetime.shutdown (0, 30.0)

def hHUP (signum, frame):
	lifetime.shutdown (3, 30.0)

class	SMTPDeliverAgent (daemon_class.DaemonClass):
	CONCURRENTS = 2
	MAX_RETRY = 3
	UNDELIVERS_KEEP_MAX = 3600
	NAME = "smtpda"

	def __init__ (self, logpath, varpath, consol, config):
		self.config = config
		daemon_class.DaemonClass.__init__ (self, logpath, varpath, consol)
		self.que = {}
		self.actives = {}
		self.last_maintern = time.time ()
		self.shutdown_in_progress = False

	def clean_shutdown_control (self, phase, time_in_this_phase):
		if phase == 1:
			self.shutdown_in_progress = True

	def run (self):
		lifetime.loop (3.0)

	def setup (self):
		self.bind_signal (hTERM, hHUP)

		val = self.config.get ("max_retry", 10)
		if val: self.MAX_RETRY = val
		val = self.config.get ("keep_days", 3)
		if val: self.UNDELIVERS_KEEP_MAX = val * 3600 * 24

		if self.config.get ("smtpserver"):
			self.log ("using default SMTP: {}".format (self.config.get ("smtpserver")), "info")
		else:
			self.log ("no default SMTP server provided", "warn")

		self.default_smtp = (
			self.config.get ("smtpserver"),
			self.config.get ("user"),
			self.config.get ("password"),
			self.config.get ("ssl") and True or False
		)

		# dummy file like object for keeping lifetime loop
		if os.name == "nt":
			select_trigger.trigger.address = ('127.9.9.9', 19998)
		select_trigger.trigger ()

		self.path_spool = os.path.join (self.varpath, "spool")
		self.log ("mail spooled in {}".format (self.path_spool), "info")
		pathtool.mkdir (self.path_spool)
		composer.Composer.SAVE_PATH = self.path_spool

		self.path_undeliver = os.path.join (self.varpath, "undeliver")
		self.log ("undelivered mail saved in {}".format (self.path_undeliver), "info")
		pathtool.mkdir (self.path_undeliver)

		lifetime.init (logger = self.logger)
		lifetime.maintern.sched (2.0, self.handle_spool)
		if os.name == "nt":
			lifetime.maintern.sched (2.0, self.maintern_shutdown_request)

	def send (self):
		if self.shutdown_in_progress:
			return

		while self.que:
			fn = self.que.popitem () [0]
			if fn in self.actives: continue

			self.actives [fn] = time.time ()
			path = os.path.join (self.path_spool, fn)
			cmps = composer.load (path)
			if cmps.get_SMTP () is None:
				cmps.set_smtp (*self.default_smtp)
			if cmps.is_SSL ():
				request = async_smtp.SMTP_SSL
			else:
				request = async_smtp.SMTP
			request (cmps, self.logger, self.when_done)
			break

	def when_done (self, cmps, code, reps):
		fn = os.path.split (cmps.get_FILENAME ()) [-1]

		if code == -250:
			cmps.remove ()
		else:
			if cmps.get_RETRYS () > self.MAX_RETRY:
				cmps.moveto (self.path_undeliver)
			else:
				cmps.moveto (self.path_spool)

		del self.actives [fn]
		self.send ()

	def maintern (self, current_time):
		for path in glob.glob (os.path.join (self.path_undeliver, "*.*")):
			mtime = os.path.getmtime (path)
			if mtime + self.UNDELIVERS_KEEP_MAX > current_time: # over a month
				try: os.remove (path)
				except: self.logger.trace ()
		self.last_maintern = time.time ()

	def handle_spool (self, current_time):
		if self.shutdown_in_progress:
			return

		if self.last_maintern + 3600 >  current_time:
			self.maintern (current_time)

		if not self.que: # previous que has priority
			for path in glob.glob (os.path.join (self.path_spool, "*.*")):
				try:
					fn = os.path.split (path)[-1]
					retrys, digest = fn.split (".")
					assert (len (digest) == 32)
					retrys = int (retrys)
				except (ValueError, AssertionError):
					continue
				except:
					self.logger.trace ()
					continue

				mtime = os.path.getmtime (path)
				if retrys > 7:
					delta = 21600
				elif retrys == 0:
					delta = 0
				else:
					delta = 4 ** retrys

				if retrys >= self.MAX_RETRY:
					self.que [fn] = None
				elif mtime + delta <= current_time and fn not in self.que and fn not in self.actives:
					self.que [fn] = None

		for i in range (self.CONCURRENTS - len (self.actives)):
			self.send ()


def usage ():
	print("""
usage:
  skitai smtpda [options...] [start|restart|stop|status]

options:
      --help
  -d: daemonize, shortcut for <start> command
      --help
      --max-retry=<int>
      --keep-days=<int>
  -s, --smtp-server=<server_addr:port>
  -u, --user=<username>
  -p, --password=<password>
      --ssl

examples:
  ex. skitai smtpda -v
	""")
	sys.exit ()


def main ():

	_default_conf = os.path.join (os.environ.get ("HOME", os.environ.get ("USERPROFILE")), ".skitai.conf")
	_fileopt = []
	if os.path.isfile (_default_conf):
		cf = confparse.ConfParse (_default_conf)
		_fileopt.extend (list ([("--" + k, v) for k, v in cf.getopt ("smtpda").items () if v not in ("", "false", "no")]))
	else:
		with open (_default_conf, "w") as f:
			f.write (DEFAULT)

	argopt = getopt.getopt (
		sys.argv[1:],
		"ds:u:p:",
		[
			"help",
			"max-retry=", "keep-days=", "server=", "user=", "password=", "ssl",
			"smtp-server="
		]
	)
	_cf = {"max-retry": 5, "keep-days": 1, "ssl": 0}
	_consol = True
	try: cmd = argopt [1][0]
	except: cmd = None
	for k, v in (_fileopt + argopt [0]):
		if k == "--help":
			usage ()
		elif k == "-d":
			cmd = "start"
		elif k == "--max-retry":
			_cf ['max-retry'] = int (v)
		elif k == "--keep-days":
			_cf ['keep-days'] = int (v)
		elif k == "-s" or k == "--smtp-server":
			_cf ['smtpserver'] = v
		elif k == "-u" or k == "--user":
			_cf ['user'] = v
		elif k == "-p" or k == "--password":
			_cf ['password'] = v
		elif k == "--ssl":
			_cf ['ssl'] = 1
	if cmd:
		_consol = False
	_logpath = os.path.join ("/var/log/skitai", SMTPDeliverAgent.NAME)
	_varpath = os.path.join ("/var/tmp/skitai", SMTPDeliverAgent.NAME)
	servicer = service.Service ("skitai/{}".format (SMTPDeliverAgent.NAME), _varpath)
	if cmd and not servicer.execute (cmd):
		return
	if not cmd and servicer.status (False):
		raise SystemError ("daemon is running")

	s = SMTPDeliverAgent (_logpath, _varpath, _consol, _cf)
	s.start ()
	sys.exit (lifetime._exit_code)


if __name__ == "__main__":
	main ()



# testing purpose WAS sync service

from . import task
from rs4 import webtest
from rs4.cbutil import tuple_cb
import random
from urllib.parse import urlparse, urlunparse
from skitai import exceptions
import xmlrpc.client
import sys
from aquests.client import synconnect

class RPCResponse:
    def __init__ (self, val):
        self.data = val

class XMLRPCServerProxy (xmlrpc.client.ServerProxy):
     def _ServerProxy__request (self, methodname, params):
        response = xmlrpc.client.ServerProxy._ServerProxy__request (self, methodname, params)
        return Result (3, RPCResponse (response))

try:
    import jsonrpclib
except ImportError:
    pass
else:
    class JSONRPCServerProxy (jsonrpclib.ServerProxy):
         def _ServerProxy__request (self, methodname, params):
            response = xjsonrpclib.ServerProxy._ServerProxy__request (self, methodname, params)
            return Result (3, RPCResponse (response))

class Result:
    def __init__ (self, status, response = None, expt = None):
        self.status = status
        self.__response = response
        self.__expt = expt

    def __getattr__ (self, attr):
        return getattr (self.__response, attr)

    @property
    def data (self):
        if isinstance (self.__response, RPCResponse):
            return self.__response.data
        elif hasattr (self.__response, "status_code"):
            ct = self.__response.headers.get ("content-type", "")
            if ct:
                if ct.startswith ("application/json"):
                    return self.__response.json ()
            return self.__response.text
        return self.__response

    def reraise (self):
        if self.status !=3 and self.__expt:
            raise self.__expt [1]

    def fetch (self, *args, **kargs):
        self.reraise ()
        return self.data

    def one (self, *args, **kargs):
        self.reraise ()
        if not self.data:
            raise exceptions.HTTPError ("410 Maybe Gone")
        elif len (self.data) != 1:
            raise exceptions.HTTPError ("409 Conflict")
        return self.data [0]


class ProtoCall (task.Task):
    def __init__ (self, cluster, *args, **kargs):
        self.cluster = cluster
        self.result = None
        self.expt = None
        self._timeout = 10
        self.handle_request (*args, **kargs)

    def get_syncon (self, uri):
        if self.cluster:
            syncon = self.cluster.get ()
        else:
            parts = urlparse (uri)
            try:
                host, port = parts [1].split (":")
            except ValueError:
                port = parts [0] == "http" and 80 or 443
                host = parts [1]
            else:
                port = int (port)
            if parts [0] == "http":
                syncon = synconnect.SynConnect ((host, port))
            else:
                syncon = synconnect.SynSSLConnect ((host, port))
            uri = urlunparse (("", "") + parts [2:])
        syncon.connect ()
        return syncon, uri

    def create_stub (self):
        syncon, uri = self.get_syncon (self.uri)
        with syncon.webtest as cli:
            if self.reqtype == "jsonrpc":
                proxy_class = JSONRPCServerProxy
            else:
                proxy_class = XMLRPCServerProxy
            return getattr (cli, self.reqtype) (uri, proxy_class)

    def handle_request (self, uri, params = None, reqtype="rpc", headers = None, auth = None, meta = None, use_cache = True, mapreduce = False, filter = None, callback = None, cache = None, timeout = 10, caller = None):
        self._mapreduce = mapreduce
        self._meta = meta
        self.uri = uri
        self.reqtype = reqtype

        syncon, uri = self.get_syncon (uri)
        syncon.set_auth (auth)
        with syncon.webtest as cli:
            req_func = getattr (cli, reqtype)
            try:
                resp = req_func (uri, headers = headers, auth = auth)
            except:
                self.expt = sys.exc_info ()
                self.result = Result (1, expt = self.expt)
            else:
                self.result = Result (3, resp)
            syncon.set_active (False)
        self.result.meta = meta or {}
        callback and callback (self.result)

    def set_callback (self, callback, reqid = None, timeout = 10):
        if reqid is not None:
            self.result.meta ["__reqid"] = reqid
        tuple_cb (self.result, callback)

    def wait (self, timeout = 10, *args, **karg):
        pass

    def _or_throw (self):
        if self.expt:
            raise exceptions.HTTPError ("700 Exception", self.expt)
        if self.result.status_code >= 300:
            raise exceptions.HTTPError ("{} {}".format (self.result.status_code, self.result.reason))
        return self.result

    def dispatch (self, *args, **kargs):
        if self._mapreduce:
            self.result = task.Results ([self.result])
        return self.result
    getwait = dispatch
    getswait = dispatch

    def dispatch_or_throw (self):
        self.dispatch ()
        return self._or_throw ()

    def commit (self, *args, **karg):
        return self._or_throw ()
    wait_or_throw = commit

    def fetch (self, timeout = 10, *args, **karg):
        self._or_throw ()
        return self.result.fetch ()

    def one (self, timeout = 10, *args, **karg):
        self._or_throw ()
        return self.result.one ()



import os, sys
import time
import tempfile
from hmac import new as hmac
import copy
from hashlib import sha1, md5
import json
import xmlrpc.client as xmlrpclib
import base64
import pickle
import random
import threading
from rs4 import pathtool, logger, jwt, deco
from aquests.protocols.smtp import composer
from aquests.protocols.http import http_date, util
from skitai import __version__, WS_EVT_OPEN, WS_EVT_CLOSE, WS_EVT_INIT, NAME, DEFAULT_BACKGROUND_TASK_TIMEOUT
from skitai import lifetime
from . import server_info
from ..corequest import tasks
from ..corequest.pth import sp_task
from ..backbone import http_response
from .triple_logger import Logger
if os.environ.get ("SKITAIENV") == "PYTEST":
    from threading import RLock
else:
    from multiprocessing import RLock
from skitai import was as the_was
import xmlrpc.client as xmlrpclib
from rs4.producers import file_producer
from .api import API

if os.environ.get ("SKITAIENV") == "PYTEST":
    from .semaps import TestSemaps as Semaps
else:
    from .semaps import Semaps

if os.name == "nt":
    TEMP_DIR =  os.path.join (tempfile.gettempdir(), "skitai-gentemp")
else:
    TEMP_DIR = "/var/tmp/skitai-gentemp"
pathtool.mkdir (TEMP_DIR)

composer.Composer.SAVE_PATH = os.path.join ("/var/tmp/skitai", "smtpda", "spool")
pathtool.mkdir (composer.Composer.SAVE_PATH)

class WASBase:
    version = __version__
    ID = None # set by skitai.was
    objects = {}
    _luwatcher = Semaps ([], "d", 256)
    lock = plock = RLock ()
    init_time = time.time ()
    cloned = False
    _process_locks = [RLock () for i in range (8)]

    @classmethod
    def get_lock (cls, name = "__main__"):
        return cls._process_locks [hash (name) % 8]

    # application friendly methods -----------------------------------------
    @classmethod
    def register (cls, name, obj):
        if hasattr (cls, name):
            raise AttributeError ("server object `%s` is already exists" % name)
        cls.objects [name] = obj
        setattr (cls, name, obj)

    @classmethod
    def unregister (cls, name):
        del cls.objects [name]
        return delattr (cls, name)

    @classmethod
    def add_handler (cls, back, handler, *args, **karg):
        h = handler (cls, *args, **karg)
        if hasattr (cls, "httpserver"):
            cls.httpserver.install_handler (h, back)
        return h

    def _clone (self):
        new_was = the_was._get (True) # get clone was
        if hasattr (self, 'env'):
            new_env = copy.copy (self.env)
            new_env ["skitai.was"] = new_was
            new_was.env = new_env
        # cloning
        for attr in ('request', 'app', 'apps', 'subapp', 'response'):
            try: val = getattr (self, attr)
            except AttributeError: pass
            else: setattr (new_was, attr, val)
        return new_was

    @property
    def timestamp (self):
        return int (time.time () * 1000)

    @property
    def uniqid (self):
        return "{}{}".format (self.timestamp, self.gentemp () [-7:])

    def __dir__ (self):
        return self.objects.keys ()

    def __str__ (self):
        return "skitai was for {}".format (threading.currentThread ())

    def in__dict__ (self, name):
        return name in self.__dict__

    # utils -----------------------------------------------
    def txnid (self):
        return "%s/%s" % (self.request.gtxid, self.request.ltxid)

    def rebuild_header (self, header, method, data = None, internal = True):
        nheader = util.normheader (header)
        if method in {"get", "delete", "post", "put", "patch", "upload"}:
            try:
                default_request_type = self.app.config.get ("default_request_type")
            except AttributeError:
                default_request_type = None
            if not default_request_type:
                default_request_type = self.DEFAULT_REQUEST_TYPE
            util.set_content_types (nheader, data, default_request_type)

        if internal:
            nheader ["X-Gtxn-Id"] = self.request.get_gtxid ()
            nheader ["X-Ltxn-Id"] = self.request.get_ltxid (1)
        else:
            nheader ["X-Requested-With"] = NAME
        return nheader

    # concurrencies ----------------------------------------------
    def Tasks (self, reqs, timeout = 10, **args):
        self.response.set_timeout (timeout)
        return tasks.Tasks (reqs, timeout, **args)

    def Future (self, req, timeout = 10, **args):
        # deprecated, use corequest.then ()
        if isinstance (req, (list, tuple)):
            raise ValueError ('Future should be single corequest')
        self.response.set_timeout (timeout)
        return tasks.Future (req, timeout, **args)

    def Futures (self, reqs, timeout = 10, **args):
        # deprecated, use was.Tasks.then ()
        if not isinstance (reqs, (list, tuple)):
            raise ValueError ('Futures should be multiple corequests')
        self.response.set_timeout (timeout)
        return tasks.Futures (reqs, timeout, **args)
    future = Future
    futures = Futures

    def Thread (self, func, *args, **kargs):
        return self.executors.create_thread (self.ID, func, *args, **kargs)

    def Process (self, func, *args, **kargs):
        return self.executors.create_process (self.ID, func, *args, **kargs)

    def Subprocess (self, cmd, timeout = DEFAULT_BACKGROUND_TASK_TIMEOUT):
        return sp_task.Task (self.ID, cmd, timeout)

    def Mask (self, data = None, _expt = None, meta = None, **attr):
        return tasks.Mask (data, _expt, self._set_was_id (meta), **attr)

    # system functions ----------------------------------------------
    def log (self, msg, category = "info", at = "app"):
        self.logger (at, msg, "%s:%s" % (category, self.txnid ()))

    def traceback (self, id = "", at = "app"):
        if not id:
            id = self.txnid ()
        self.logger.trace (at, id)

    def render_ei (self, exc_info, format = 0):
        return http_response.catch (format, exc_info)

    @property
    def tempfile (self):
        return self.gentemp ()

    def gentemp (self):
        return os.path.join (TEMP_DIR, next (tempfile._get_candidate_names()))

    def email (self, subject, snd, rcpt):
        return composer.Composer (subject, snd, rcpt)

    # -- only allpy to current worker process
    def status (self, flt = None, fancy = True):
        return server_info.make (self, flt, fancy)

    def restart (self, timeout = 0):
        lifetime.shutdown (3, timeout)

    def shutdown (self, timeout = 0):
        lifetime.shutdown (0, timeout)

    # inter-processes communication ------------------------------
    def setlu (self, name, *args, **karg):
        self._luwatcher.set (name, time.time (), karg.get ("x_ignore", False))
        self.broadcast (name, *args, **karg)

    def getlu (self, *names):
        mtimes = []
        for name in names:
            mtime = self._luwatcher.get (name, self.init_time)
            mtimes.append (mtime)
        return max (mtimes)

    def setgs (self, name, val):
        assert isinstance (val, int), "global state must be integer"
        self._luwatcher.set (name, val)
        self.broadcast (name, val)

    def getgs (self, name, default = 0):
        assert isinstance (default, int), "global state must be integer"
        val = self._luwatcher.get (name, default)
        return int (val)

    def incgs (self, name, delta = 1):
        val = self.getgs (name) + delta
        self.setgs (name, val)

    # websokcet / http2 ------------------------------------------
    def push (self, uri):
        self.request.response.push_promise (uri)

    def wsconfig (self, spec, timeout = 60, encoding = "text", session = None):
        # other than atila, encoding became varname
        self.env ["websocket.config"] = (spec, timeout, encoding, session)
        return ""

    def wsinit (self):
        return self.env.get ('websocket.event') == WS_EVT_INIT

    def wsopened (self):
        return self.env.get ('websocket.event') == WS_EVT_OPEN

    def wsclosed (self):
        return self.env.get ('websocket.event') == WS_EVT_CLOSE

    def wshasevent (self):
        return self.env.get ('websocket.event')

    def wsclient (self):
        return self.env.get ('websocket.client')

    # will be deprecated --------------------------------------------------
    def togrpc (self, obj):
        return obj.SerializeToString ()

    def fromgrpc (self, message, obj):
        return message.ParseFromString (obj)

    def tojson (self, obj):
        try:
            encoder = self.app.config.get ('JSON_ENCODER')
        except AttributeError:
            encoder = None
        return json.dumps (obj, cls = encoder)

    def toxml (self, obj):
        return xmlrpclib.dumps (obj, methodresponse = False, allow_none = True, encoding = "utf8")

    def fromjson (self, obj):
        if type (obj) is bytes:
            obj = obj.decode ('utf8')
        return json.loads (obj)

    def fromxml (self, obj, use_datetime = 0):
        return xmlrpclib.loads (obj)

    def fstream (self, path, mimetype = 'application/octet-stream'):
        self.response.set_header ('Content-Type',  mimetype)
        self.response.set_header ('Content-Length', str (os.path.getsize (path)))
        return file_producer (open (path, "rb"))

    def jstream (self, obj, key = None):
        self.response.set_header ("Content-Type", "application/json")
        if key:
            # for single skeleton data is not dict
            return self.tojson ({key: obj})
        else:
            return self.tojson (obj)

    def xstream (self, obj, use_datetime = 0):
        self.response.set_header ("Content-Type", "text/xml")
        return self.toxml (obj, use_datetime)

    def gstream (self, obj):
        self.response.set_header ("Content-Type", "application/grpc")
        return self.togrpc (obj)
